<?php

// +----------------------------------------------------------------------+
// | jPHPsetistats.php                                                    |
// |   a user Seti@home statistics retriever (using XML).                 |
// +----------------------------------------------------------------------+
// |        jPHPsetistats project                                         |
// | coded in PHP version 4 in 2004 by Joel Thieffry & the zeRezo group   |
// +----------------------------------------------------------------------+
// |  This source file is subject to the Lesser GNU Public Licence, which |
// | text can be found at http://www.gnu.org/licenses/lgpl.txt .          |
// +----------------------------------------------------------------------+
// | Authors: Joel Thieffry <jo@zerezo.com>                               |
// +----------------------------------------------------------------------+
/*
 * TODO : generate HTML tab with statistics, like pstats.
 * TODO : write first to temporary stat file, then rename/move that file to the target.
          Or use ignore_user_abort(true); or fopen ($tmp, 'x');
 * MAYDO : gzipping of local data files. NO.
 * MAYDO : gather access stats from setipng.php script.
 * MAYDO : stats file locking.
 * 20040731 jo : new version: 0.04. Because the previous added feature was all I needed to safely release.
 * 20041117 jo : added email encoding.
 * 20040731 jo : new version: 0.03.
 * 20040829 jo : cache cleaning (when exiting).
 * 20040829 jo : when seti@home server is down, use lastest local statistics (if not flushed nor flushable).
 * 20040829 jo : when getting an image and an error occurs, generate the image with the error text.
 * 20040801 jo : added a function to get the last certificate value.
 * 20040801 jo : added a function to get the hidden certificate URL.
 * 20040801 jo : corrected certificate URLs (replacing "+" with "&").
 * 20040731 jo : new version: 0.02.
 * 20040731 jo : added options to control access by email list.
 * 20040731 jo : added an option to deactivate the local cache.
 * 20040731 jo : added quick'n'dirty replacement function for PHP5 file_put_contents.
 * 20040729 jo : added quick email checking (to avoid the leak when writing data).
 * 20040729 jo : added local data cache.
 * 20040729 jo : initial version: 0.01.
 */

/////
// Some constants.
// ///

  /** jPHPsetistats version number. */
  define("JPHPSETISTATS_VERSION_NUMBER", "0.04");
  /** The prefix to the link to seti@home user stats in XML. */
  define("JPHPSETISTATS_SETI_USER_STATS_XML_PREFIX", "http://setiathome2.ssl.berkeley.edu/fcgi-bin/fcgi?cmd=user_xml&email=");
  /** The text of the error, if 'not parsed yet'. */
  define("JPHPSETISTATS_ERROR_NOT_PARSED_YET", "Not parsed yet.");
  /** A default email address. */
  define("JPHPSETISTATS_DEFAULT_EMAILURL", "royale-AT-rezo-DOT-dyndns-DOT-org");
  /**
    * Minimum interval in seconds between two consecutive retreival from the seti website.
    * Set it zero or a negative number to always retrieve.
    */
  define("JPHPSETISTATS_REFRESH_MINIMUM_INTERVAL_IN_SECONDS", 60*15); // set to 15 minutes.
  /**
    * Minimum interval in seconds before flushing data from cache.
    * Set it zero or a negative number to never flush.
    */
  define("JPHPSETISTATS_FLUSH_MINIMUM_INTERVAL_IN_SECONDS", 60*60*24*15); // set to 15 days.
  /** Path to where save the retrieved data. Can be relative (fortunately!). */
  define("JPHPSETISTATS_CACHE_PATH", "./statscache");
  /** Activate or not the local cache (data stored in JPHPSETISTATS_CACHE_PATH, see above). */
  define("JPHPSETISTATS_USE_LOCAL_CACHE", TRUE);
  /** If TRUE, restricts email adresses to the ones in $jPHPsetistats_EmailTab. */
  define("JPHPSETISTATS_RESTRICT_EMAIL", FALSE);

  /** The array of explicitly allowed email URLs. */
  $jPHPsetistats_EmailTab = array("royale-AT-rezo-DOT-dyndns-DOT-org",
                                  "joet3-AT-ifrance-DOT-com");

  /**
   * Seti@Home user stats retreiver class.
   */
  class jPHPsetistats {

/////
// Internal variables.
// ///

    /** Email of the user, needed to get the user stats from Seti@Home website. */
    var $emailURL;
    /** Description of the error, if occured, otherwise null. */
    var $errorText;
    /** For parsing usage only. */
    var $lastElement;
    /** The localpath to the cache, or null if not using local cache (or can't use). */
    var $localCachePath;
    /** The name of the local file that will hold data of the user, or null if not using local cache (or can't use). */
    var $localFilename;

/////
// Data variable that old valuable information.
// ///

    /** User name. */
    var $name;
    /** User webpage URL (optional). */
    var $userURL;
    /** Number of results returned by user. */
    var $numresults;
    /** Total cpu time compiled by user. */
    var $cputime;
    /** Average cpu time per returned result. */
    var $avecpu;
    /** Average number of results returned per day. */
    var $resultsperday;
    /** Time last result was returned. */
    var $lastresulttime;
    /** User registration date. */
    var $regdate;
    /** Length of time user has been registered. */
    var $usertime;
    /** Link to user profile (optional). */
    var $userprofileURL;
    /** Name of user's group (optional). */
    var $group;
    /** Link to user's group (optional). */
    var $groupURL;
    /* Link to group for which user is founder (optional, may be more than one). */
    var $founder;
    /** Rank within total user population. */
    var $rank;
    /** Total user population. */
    var $ranktotalusers;
    /** Number of other users who share the same rank. */
    var $num_samerank;
    /** Percent of the user population ranked ahead of the user. */
    var $top_rankpct;
    /** Links to milestone certificate (optional, may be more than one). */
    var $certURLTab;
    /** Link to the hidden certificate. */
    var $hiddenCertURL;
    /** Value of the last certificate, in workunits. */
    var $lastCertValue;
    
    /**
      * Class constructor.
      * @param $_emailURL The user email. EXACT spelling and casing !
      */
    function &jPHPsetistats($_emailURL = null) {
      if (is_null($_emailURL))
        $this->emailURL = $this->emailDecode(JPHPSETISTATS_DEFAULT_EMAILURL);
      else
        $this->emailURL = strtolower($_emailURL);
      $this->emailURL = strtolower($this->emailURL);
      $this->errorText = JPHPSETISTATS_ERROR_NOT_PARSED_YET;
      $this->localCachePath = $this->localFilename = null;
      $this->name = $this->userURL = $this->numresults = $this->cputime
                  = $this->avecpu = $this->resultsperday = $this->lastresulttime
                  = $this->regdate = $this->usertime = $this->userprofileURL
                  = $this->group = $this->groupURL = $this->founder
                  = $this->rank = $this->ranktotalusers = $this->num_samerank
                  = $this->top_rankpct = "";
      $this->certURLTab = array();
      $this->hiddenCertURL = "http://setiathome.ssl.berkeley.edu/cgi-bin/cgi2?email="
                       .$this->emailURL."&cmd=print_cert&certnum=0&size=0";
      $this->lastCertValue = 0;
      
      /* We may use MD5 to avoid the potential leak, but it's less readable. */
      if (preg_match('/[^0-9a-z@_.\-]/', $this->emailURL))
        $this->errorText = "Incorrect email address";
      else if (JPHPSETISTATS_USE_LOCAL_CACHE)
      {
        $this->localCachePath = JPHPSETISTATS_CACHE_PATH;
        if ($this->localCachePath == "")
          $this->localCachePath = "./";
        else {
          $endChar = $this->localCachePath[strlen($this->localCachePath) - 1];
          if (($endChar != "/") && ($endChar != "\\"))
            $this->localCachePath .= "/";
        }
        if (!file_exists($this->localCachePath) || !is_dir($this->localCachePath)) {
          //echo "Local cache path \"$this->localCachePath\" doesn't work.<br />\n";
          $this->localCachePath = null;
        } else
          $this->localFilename = $this->localCachePath.$this->emailURL.".xml";
      }
    }
    
    /**
      * Get a description of the occured error.
      * @return The error text, or null;
      */
    function getError() {
      return $this->errorText;
    }
    
    /**
      * Try to get the stats from the disk.
      * @param _isMinimumInterval TRUE (by default) if testing date with minimum refresh interval, or FALSE to test it with minimum cache flush interval.
      * @return the content if loaded successfully, otherwise null
      *  (but no error reported, because it's not critical).
      */
    function getStatsFromDisk($_isMinimumInterval = TRUE) {
      if (!JPHPSETISTATS_USE_LOCAL_CACHE)
        return null;

      if (!file_exists($this->localFilename)) {
        //echo ("Local file \"$this->localFilename\" not found.<br />\n");
        return null;
      }
      $interval = time() - filemtime($this->localFilename);
      if ($_isMinimumInterval && ($interval >= ($_isMinimumInterval?JPHPSETISTATS_REFRESH_MINIMUM_INTERVAL_IN_SECONDS:JPHPSETISTATS_FLUSH_MINIMUM_INTERVAL_IN_SECONDS))) {
        //echo ("Refresh time interval elapsed : ".$interval.".<br />\n");
        return null;
      }
      if (!($fileContent = @file_get_contents($this->localFilename))) {
        //echo ("Error loading local file.<br />\n");
        return null;
      }
      //echo ("Local file read.<br />\n");
      return $fileContent;
    }
    
    /**
      * Try to retrieve the user stats from Seti@Home website.
      * @return the content if retrieved successfully, otherwise null (along with an error).
      */
    function getStatsFromWebsite() {
      //$filename = "setistats.xml"; // For testing.
      $filename = JPHPSETISTATS_SETI_USER_STATS_XML_PREFIX.$this->emailURL;

      if (strncmp(strtolower($filename), "http://", 7) == 0)
        if (@ini_get('allow_url_fopen') == 0) {
          $this->errorText = "You need to set 'allow_url_fopen' to 'On' in your php.ini directives.";
          return null;
        }

      if (!($fileContent = @file_get_contents($filename))) {
        $this->errorText = "Can't open XML file \"$filename\"";
        return null;
      }
      return $fileContent;
    }

    /**
      * Parse the xml data to get all stats.
      * @param $_data The XML data string.
      * @return TRUE if parsed successfully, otherwise FALSE.
      */
    function parseData($_data)
    {
      if (is_null($_data))
        return FALSE;
      $xml_parser = xml_parser_create("UTF-8");
      xml_parser_set_option($xml_parser, XML_OPTION_CASE_FOLDING, FALSE);
      xml_set_element_handler($xml_parser, array(&$this,"startElement"),
                                           array(&$this,"endElement"));
      xml_set_character_data_handler($xml_parser, array(&$this,"charData"));

      $this->errorText = null;
      if (!xml_parse($xml_parser, $_data, TRUE)) {
        $this->errorText = "XML error: ".xml_error_string(xml_get_error_code($xml_parser).
                           " in line ".xml_get_current_line_number($xml_parser)).".";
        xml_parser_free($xml_parser);
        return FALSE;
      }

      if ($this->certURLTab) {
        $explodedURL = explode("&", $this->certURLTab[0]);
        $this->lastCertValue = substr($explodedURL[2], 8);
      } //else $this->lastCertValue = 0;

      xml_parser_free($xml_parser);
      return TRUE;
    }

    function cleanCache()
    {
      if (JPHPSETISTATS_USE_LOCAL_CACHE && !is_null($this->localCachePath))
      {
        //echo "Flushing.<br />\n";
        $dirHandle = opendir($this->localCachePath);
        while (FALSE !== ($filename = readdir($dirHandle))) {
          $fullFilename = $this->localCachePath.$filename;
          //echo "Found \"$filename\"<br />\";
          if ( is_file($fullFilename) &&
               (strlen($filename) >= 6) &&
               (strcmp(".xml", substr($filename, -4)) == 0) &&
               (time() - filemtime($fullFilename) > JPHPSETISTATS_FLUSH_MINIMUM_INTERVAL_IN_SECONDS) ) {
            //echo "Flushing \"$fullFilename\".<br />\n";
            @unlink($fullFilename);
          }
        }
      }
    }

    /**
      * Retrieve the user stats.
      * @return TRUE if successfully retrieved, otherwise FALSE (along with an error).
      */
    function getStats() {
      global $jPHPsetistats_EmailTab;
      
      if ($this->errorText !== JPHPSETISTATS_ERROR_NOT_PARSED_YET)
        return FALSE;

      if (JPHPSETISTATS_RESTRICT_EMAIL) {
        foreach ($jPHPsetistats_EmailTab as $key => $email)
          $jPHPsetistats_EmailTab[$key] = $this->emailDecode($email);
        if (! in_array($this->emailURL, $jPHPsetistats_EmailTab)) {
          $this->errorText = "Email not allowed because of restriction policy.";
          return FALSE;
        }
      }

      $isNotLocalFile = is_null($fileContent = $this->getStatsFromDisk());
      if ($isNotLocalFile)
        if (is_null($fileContent = $this->getStatsFromWebsite()))
          if ($isNotLocalFile = is_null($fileContent = $this->getStatsFromDisk(FALSE)))
            return FALSE;
          
      //echo "<pre>Content=$fileContent</pre><br />\n";

      if (! $this->parseData($fileContent))
        return FALSE;

      $this->cleanCache();

      if (JPHPSETISTATS_USE_LOCAL_CACHE && $isNotLocalFile && !is_null($this->localFilename))
      {
        //echo "Trying to write : ";
        $fileRes = @$this->file_put_contents($this->localFilename, $fileContent);
        /*if (!$fileRes)
          echo "missed !<br />\n";
        else
          echo "OK ;)<br />\n";
        */
      }
      
      return TRUE;
    }

/////
// Parser functions.
// ///

    /**
      * Write a string into a new file.
      * @param $_filename The file name.
      * @param $_data The string data to be written.
      * @return The amount of written bytes, or FALSE in case of error.
      */
    function file_put_contents($_filename, &$_data) {
      if (function_exists("file_put_contents"))
        return file_put_contents($_filename, $_data);
      else {
        $file = fopen($_filename, 'w');
        if (!$file)
          return FALSE;
        else {
          $nbBytes = fwrite($file, $_data, strlen($_data));
          fclose($file);
          return $nbBytes;
        }
      }
    }

    /** A callback to the xml parser. */
    function startElement($_parser, $_name, $_attrTab)
    {
      if ($_name === "a") {
        switch ($this->lastElement) {
          case "name":
            @$this->userURL = $_attrTab["href"];
            break;
          case "userprofile":
            @$this->userprofileURL = $_attrTab["href"];
            break;
          case "group":
            @$this->groupURL = $_attrTab["href"];
            break;
          case "cert":
            @$this->certURLTab[] = str_replace("+", "&", $_attrTab["href"]);
            break;
        }
      } else
        $this->lastElement = $_name;
    }

    /** A callback to the xml parser. */
    function endElement($_parser, $_name)
    {
      $this->lastElement = null;
    }

    /** A callback to the xml parser. */
    function charData($_parser, $_data)
    {
      switch ($this->lastElement) {
        case "name":
          @$this->name = $_data;
          break;
        case "numresults":
          @$this->numresults = $_data;
          break;
        case "cputime":
          @$this->cputime = $_data;
          break;
        case "avecpu":
          @$this->avecpu = $_data;
          break;
        case "resultsperday":
          @$this->resultsperday = $_data;
          break;
        case "lastresulttime":
          @$this->lastresulttime = $_data;
          break;
        case "regdate":
          @$this->regdate = $_data;
          break;
        case "usertime":
          @$this->usertime = $_data;
          break;
        case "group":
          @$this->group = $_data;
          break;
        case "founder":
          @$this->founder = $_data;
          break;
        case "rank":
          @$this->rank = $_data;
          break;
        case "ranktotalusers":
          @$this->ranktotalusers = $_data;
          break;
        case "num_samerank":
          @$this->num_samerank = $_data;
          break;
        case "top_rankpct":
          @$this->top_rankpct = $_data;
          break;
      }
    }

/////
// Image generators.
// ///

    /**
      * Generate an image only with specified text.
      * @param _text The text to be written in the image.
      * @return The generated image.
      */
    function createTextImage($_text)
    {
      if (empty($_text))
        $_text = " ";
      $img = ImageCreateTrueColor(strlen($_text)*7, 14);
      $white = ImageColorAllocate($img, 255, 255, 255);
      ImageFill($img, 0, 0, $white);
      ImageColorTransparent($img, $white);
      $black = ImageColorAllocate($img, 0, 0, 0);
      ImageString($img, 3, 0, 0, $_text, $black);
      return $img;
    }

    /**
      * Generate a simple image for signature.
      * @return The generated image, with stats or with occured error text.
      */
    function getStatImage()
    {
      if ($this->errorText === JPHPSETISTATS_ERROR_NOT_PARSED_YET)
        $this->getStats();
      if (!is_null($this->errorText))
        return $this->createTextImage($this->errorText);
        
      $text = 'Seti@Home stats for '.$this->name.' : '.$this->numresults.' units / '.trim($this->cputime).'.';
      return $this->createTextImage($text);
    }

    /*
     * Send a generated image from signature image to the browser.
     * @param _format The format to generate the image in ("PNG", "GIF", "JPG" or "WBMP").
     */
    function showImage($_format) {
      $image = $this->getStatImage();
      if (is_null($image)) {
        if (is_null($this->errorText))
          $this->errorText = "Error while rendering the image.";
        return;
      }
      $_format = strtoupper($_format);
      switch ($_format) {
        case 'PNG':
          if (!(imagetypes() & IMG_PNG)) {
            $this->errorText = "This server doesn't support PNG image format generation.";
            return;
          }
          Header("Content-type: image/png");
          ImagePNG($image);
          break;
        case 'JPG':
          if (!(imagetypes() & IMG_JPEG)) {
            $this->errorText = "This server doesn't support PNG image format generation.";
            return;
          }
          Header("Content-type: image/jpeg");
          ImageJPEG($image);
          break;
        case 'WBMP':
          if (!(imagetypes() & IMG_WBMP)) {
            $this->errorText = "This server doesn't support PNG image format generation.";
            return;
          }
          Header("Content-type: Content-type: image/vnd.wap.wbmp");  // to check
          ImageWBMP($image);
          break;
        default: // GIF is default
          if (!(imagetypes() & IMG_GIF)) {
            $this->errorText = "This server doesn't support PNG image format generation.";
            return;
          }
          Header("Content-type: image/gif");
          ImageGIF($image);
      }
    }

    /*
     * Send a generated PNG from signature image to the browser.
     */
    function showPNG() { $this->showImage('PNG'); }

    /*
     * Send a generated GIF from signature image to the browser.
     */
    function showGIF() { $this->showImage('GIF'); }

    /*
     * Send a generated JPG from signature image to the browser.
     */
    function showJPG() { $this->showImage('JPG'); }

    /*
     * Send a generated WBMP from signature image to the browser.
     */
    function showWBMP() { $this->showImage('WBMP'); }

    /*
     * Encode an email address to prevent spam on it.
     * Basically trancode "@" and "." to "-AT-" and "-DOT-".
     * @param _email The email string to be encoded.
     * @return The encoded email string.
     */
    function emailEncode($_email) {
      $encodeTrTab = array("@" => "-AT-", "." => "-DOT-");
      if (is_null($_email))
        return $_email;
      return strtr($_email, $encodeTrTab);
    }

    /*
     * Decode an email address encoded with emailEncode().
     * @param _email The email string to be decoded.
     * @return The decoded email string.
     */
    function emailDecode($_email) {
      $decodeTrTab = array("-AT-" => "@", "-at-" => "@", "-DOT-" => ".",
                           "-dot-" => ".", "-NOSPAM-" => "", "-nospam-" => "");
      if (is_null($_email))
        return $_email;
      return strtr($_email, $decodeTrTab);
    }

  }
?>
